#!/usr/bin/python

"""
Turns plugin.gba + file.plugin into:
	For Multiboot:
		[fakeromheader (with new_jump)]
		[compression_type]
		[plugin.gba]
		[[reset code]]
		[[font]]
		[[fontpal]]
		[[background characters]]
		[[background palette]]
		[[background tile map]]
		[pogoshell-simulator stub]
		[[arguments array]]
		[file.plugin]
		[[padding]]
		[[fake root]]

	For non-multiboot:
		[new_jump]
		[plugin.gba (minus old jump)]
		[[reset code]]
		[[font]]
		[[fontpal]]
		[[background characters]]
		[[background palette]]
		[[background tile map]]
		[pogoshell-simulator stub]
		[[arguments array]]
		[file.plugin's]
		[[padding]]
		[[fake root]]

	Patching includes:
		computing new_jump
		for non-multiboot, noting old jump and storing value in pogoshell-simulator stub
		noting various addresses and storing values in pogoshell-simulator stub
		create arguments block with file data for menuing and constructing final passed arguments
		for pseudo-filesystem, creating a multi-entry pogoshell filesystem referencing plugin.bin and file1.plugin ... filen.plugin
		various padding to make sure nearly everything is word aligned 

	Execution of plugin.gba + ... + file.plugin is generally:
		new_jump into pogoshell-simulator stub
		look for magic key
		clear almost all of ewram
		copy magic key to special place in ewram
		if multiboot, copy plugin.gba to ewram
		if argument_count is greater than one, select argument
		copy arguments to ewram
		if multiboot, jump to ewram
		if non-multiboot, jump to old jump address

	If magic key is found:
		then this is the second run through (probably a swi 0x00)
		so, call reset code

	Todo:
		figure out why drsms.bin doesn't work
		probably choose a better location for the magic key

	Not doable:
		patching non-multiboot plugins without a romheader of any kind (ex nsf.bin*)

	*Note: nsf.bin is multibootable; it's the reason for the fakeromheader
"""

import os, sys, struct, getopt
from glob import glob
from sys import argv
from stat import *

MULTIBOOT_RAW = 1
MULTIBOOT_LZ77 = 2
MULTIBOOT_APACK = 3

def readfile(name):
	try:
		fd = open(name, "rb")
		contents = fd.read()
		fd.close()
	except IOError:
		print "Error reading", name
		sys.exit(1)
	return contents

def writefile(name, contents):
	try:
		fd = open(name, "wb")
		fd.write(contents)
		fd.close()
	except IOError:
		print "Error writing", name
		sys.exit(1)

def make_arguments(args, offset):
	offset += len(args)*4*3
	names = []
	array = ""
	post_array = ""
	for a in args:
		file = readfile(a)
		filename = make_valid_name(a)
		if filename in names:
			i = 0
			newname = filename
			while newname in names:
				newname = make_unique(filename,i)
				i += 1
			filename = newname
		names.append(filename)
		filename += "\0" * ((4 - (len(filename)&3))&3)
		filesize = len(file)
		file += "\0" * ((4 - (len(file)&3))&3)
		array += struct.pack("<3I", filesize, offset + len(filename), offset)
		post_array += filename + file
		offset += len(filename) + len(file)
	return array + post_array

def make_unique(name,number):
	name = name[:-1]
	strnumber = "%04d" % number
	extension = name.rfind(".")
	if len(name) > 31-4:
		if extension != -1:
			new_name = name[:31-4-len(name)+extension] + strnumber + name[extension:]
		else:
			new_name = strnumber + name[extension:extension+31-4]
	else:
		if extension != -1:
			new_name = name[:extension] + strnumber + name[extension:]
		else:
			new_name = strnumber + name
	return new_name + "\0"

def make_valid_name(name):
	slash = name.rfind("/")
	slash2 = name.rfind("\\")
	if slash2 > slash:
		slash = slash2
	name = name[slash+1:]
	new_name = name
	extension = name.rfind(".")
	if extension != -1:
		if len(name)-extension > 31:
			new_name = name[extension:extension+31]
		elif len(name) > 31:
			new_name = name[:31-len(name)+extension] + name[extension:]
	else:
		if len(name) > 31:
			new_name = name[0:31]
	tmp_name = ""
	for f in new_name:
		if not 32 <= ord(f) <= 127:
			tmp_name += "_"
		else:
			tmp_name += f
	new_name = tmp_name + "\0"
	return new_name

def create_fake_filesystem(offset, whole, cmdname, cmdsize, cmdaddress, argcount, arguments):
	filecount = argcount + 1
	filesystem = struct.pack("<II", 0xFAB0BABE, (filecount*(32+4+4)) | 0x80000000)
	filesystem += struct.pack("<31scIi", cmdname, "\0", cmdsize, cmdaddress - offset)
	for i in range(0, argcount*12, 12):
		filesize, fileoffset, filenameoffset = struct.unpack("<3I", arguments[i:i+12])
		filename = whole[filenameoffset-0x08000000:fileoffset-0x08000000]
		#print filename, "%08x %08x" % (filesize, fileoffset)
		filesystem += struct.pack("<31scIi", filename, "\0", filesize, fileoffset - offset)
	return filesystem

#def splitbackground(background_bin):
#	"""
#	 format == [int32 header][int32 count][int32 size][char[24] title]
#             [int32 characters_size][char[characters_size] characters]
#						 [uint16[16] palette][char[size-characters_size-24-32] map]
#	"""
#	marker, count, size = struct.unpack("<3I", background_bin[0:12])
#	print ("%x" % marker), count, size, len(background_bin)
#	if marker != 0x789B4987 or count != 1 or size + 8 != len(background_bin):
#		print "Invalid background"
#		sys.exit(1)
#	characters_size = struct.unpack("<I", background_bin[36:40])[0]
#	characters = background_bin[40:40+characters_size-4]
#	palette = background_bin[40+characters_size-4:40+characters_size+32-4]
#	map = background_bin[40+characters_size+32-4:]
#	return characters, palette, map

def usage(name):
	print "Usage: %s [-fmzas] output.gba plugin.bin file1.plugin [... filen.plugin]" % name
	sys.exit(0)

if __name__ == "__main__":
	fake = 0
	multiboot = 0
	sort_args = 0

	try:
		opts, args = getopt.gnu_getopt(argv[1:], "fmzas", [])
	except getopt.GetoptError:
		usage(argv[0])

	for o, a in opts:
		if o in ("-f"):
			fake = 1
		if o in ("-m"):
			multiboot = MULTIBOOT_RAW
		if o in ("-z"):
			multiboot = MULTIBOOT_LZ77
		if o in ("-a"):
			multiboot = MULTIBOOT_APACK
		if o in ("-s"):
			sort_args = 1

	if sys.platform == "win32":
		tmp = []
		for a in args:
			expanded = glob(a)
			if len(expanded) == 0:
				tmp.append(a)
			else:
				tmp.extend(expanded)
		args = tmp

	if len(args) < 3:
		usage(argv[0])

	if sort_args:
		tmp0 = args[0:2]
		tmp1 = args[2:]
		tmp1.sort()
		tmp0.extend(tmp1)
		args = tmp0

	make_arguments_bin = readfile("make_arguments.bin")
	variables_offset = struct.unpack("<I", make_arguments_bin[0:4])[0] - 4
	make_arguments_bin = make_arguments_bin[4:]

	#variables
	#reset_code:
	#return_point:
	#plugin_cputfastset:
	#font:
	#fontpal:
	#background_char:
	#background_pal:
	#background_map:
	#cmdname:
	#argument_count:
	#arguments_array:

	reset_bin = readfile("reset.bin")
	romheader_bin = readfile("romheader.bin")
	
	plugin = readfile(args[1])
	cmdsize = len(plugin)
	plugin += "\0" * ((4 - (len(plugin)&3))&3)

	offset = 0x08000000 + len(plugin)
	if multiboot:
		offset += len(romheader_bin) + 4
		cmdaddress = 0x08000000 + len(romheader_bin) + 4
	else:
		#fix romheader
		#copy nintendo logo in and fixed value 96h and device type
		plugin = plugin[0:0x4] + romheader_bin[0x4:0xa0] + plugin[0xa0:0xb2] + "\x96" + plugin[0xb3] + "\x80" + plugin[0xb5:]
		#compute checksum
		chk = 0
		for i in range(0xa0, 0xbd):
			chk += struct.unpack("B", plugin[i])[0]
		chk = (0x100 - ((chk + 0x19) & 0xff)) & 0xff
		plugin = plugin[0:0xbd] + struct.pack("B", chk) + plugin[0xbe:]
		cmdaddress = 0x08000000

	argument_count = len(args) - 2

	plugin_cpufastset = (cmdsize+31)/32*8

	reset_address = offset
	offset += len(reset_bin)

	font_address = 0
	fontpal_address = 0
	background_char_address = 0
	background_pal_address = 0
	background_map_address = 0

	bulk_bin = ""
	if argument_count > 1:
		font_bin = readfile("font.lz77")
		fontpal_bin = readfile("fontpal.lz77")
		#background_bin = readfile("background.bin")
		#background_char_bin, background_pal_bin, background_map_bin = splitbackground(background_bin)
		background_char_bin = readfile("background.img.bin")
		background_pal_bin = readfile("background.pal.bin")
		background_map_bin = readfile("background.map.bin")

		bulk_bin = font_bin + fontpal_bin + background_char_bin + background_pal_bin + background_map_bin

		font_address = offset
		offset += len(font_bin)

		fontpal_address = offset
		offset += len(fontpal_bin)

		background_char_address = offset
		offset += len(background_char_bin)

		background_pal_address = offset
		offset += len(background_pal_bin)

		background_map_address = offset
		offset += len(background_map_bin)

	if multiboot:
		return_address = 0x02000000
	else:
		return_address = struct.unpack("<I", plugin[0:4])[0]
		if (return_address >> 24) != 0xEA:
			print "Unable to patch"
			sys.exit(2)
		return_address = return_address & 0x00FFFFFF
		return_address = 0x08000008 + return_address * 4

	new_jump_address = offset
	offset += len(make_arguments_bin)

	new_jump_address = 0xEA000000 | (new_jump_address - 0x08000008)/4

	print "%i %08x %08x" % (offset, return_address, new_jump_address)

	if multiboot:
		plugin = struct.pack("<I", new_jump_address) + romheader_bin[4:] + struct.pack("<I", multiboot - 1) + plugin
	else:
		plugin = struct.pack("<I", new_jump_address) + plugin[4:]
	
	cmdname = "/" + make_valid_name(args[1])
	cmdname += "\0" * ((4 - (len(cmdname)&3))&3)

	cmdname_address = offset
	offset += len(cmdname)

	arguments_array_address = offset
	arguments = make_arguments(args[2:], offset)

	variables = struct.pack("<11I", reset_address, return_address, plugin_cpufastset, font_address, fontpal_address, background_char_address, background_pal_address, background_map_address, cmdname_address, argument_count, arguments_array_address)
	make_arguments_bin = make_arguments_bin[0:variables_offset] + variables + make_arguments_bin[variables_offset + len(variables):]

	whole = plugin + reset_bin + bulk_bin + make_arguments_bin + cmdname + arguments

	if fake:
		whole += "\0" * ((32768 - (len(whole)&32767))&32767)
		offset = len(whole) + 0x08000000 + 8
		whole += create_fake_filesystem(offset, whole, cmdname[1:32], cmdsize, cmdaddress, len(args)-2, arguments)

	writefile(args[0], whole)
